/*

*/

#include <stdio.h>     /* gestione generica dell'I/O */
#include <stdlib.h>    /* librerie standard */
#include <graphics.h>  /* BGI per tutte le funzionalità grafiche */
#include <limits.h>    /* limiti del C (MAX_INT,...)  */

/* definizione delle costanti */
#define NUM_MAX_PUNTI 20
#define DELTAX 25
#define DELTAY 25
#define COLOREDATI RED
#define COLOREASSI BROWN
#define COLORESFONDO WHITE
#define COLOREPUNTOMEDIO CYAN
#define COLORERETTAREGRESSIONE LIGHTRED
#define COLORERETTACOSTI GREEN
#define COLORERETTARICAVI BLUE

/* dati di prova per non dover inserire ogni volta */
float DATIX [7]  = { 3,-5, 2, 4, 9};
float DATIY [7]  = { 3,-4, 3, 6, 6};

/* Struttura dati per la gestione dei punti nel piano */
typedef struct s_punto {
    float x;
    float y;
} punto;

/* Risoluzione dello schermo (ampiezza e altezza in pixel) */
int scrW, scrH;
/* valori minimo e massimo dei dati rilevati sull'asse X */
float minX = -1.0;
float maxX = INT_MIN;
/* valori minimo e massimo dei dati rilevati sull'asse Y */
float minY = -1.0;
float maxY = INT_MIN;

/* Funzione per acquisire un dato rilevato:
   d: <= qui viene salvato il dato rilevato
   idx: => numero di rilevazione
   return : intero che indica se ci sono ancora dati da rilevare
*/
int acquisisciDato (punto* d, int idx){
    int ancora=0;
    float x, y;
    printf("Inserisci la rilevazione numero %d relativa alla variabile indipendente x\n", idx);
    scanf("%f", &x);
    printf("Inserisci la rilevazione numero %d relativa alla variabile dipendente y\n", idx);
    scanf("%f", &y);
    d[idx].x = x;
    d[idx].y = y;
    printf("Ci sono altri dati da immettere? Digitare 1 se si, 0 altrimenti\n");
    scanf("%d",&ancora);
    return ancora;
}

/* Funzione di debug per acquisire un dato rilevato:
   In questo modo non si è costretti a inserire i dati ogni volta.
   d: <= qui viene salvato il dato rilevato
   idx: => numero di rilevazione
   return : intero che indica se ci sono ancora dati da rilevare
*/
int acquisisciDato2 (punto* d, int idx){
    d[idx].x = DATIX[idx];
    d[idx].y = DATIY[idx];
    return idx<6;
}

/* Funzione per la conversione di una coordinata x nel piano cartesiano in
   una coordinata x di schermo. Usa le variabili globali
   - maxX: massima coordinata X rilevata; si usa per riscalare la vista in modo da
           sfruttare al massimo lo schermo
   - minX: come sopra
   -scrW: ampiezza in pixel dello schermo
   x : => coordinata da trasformare
   return: coordinata a schermo
*/
int xToScreen (float x){
    float xMin = min (-1.0, minX);
    return  DELTAX + (x-xMin)*(scrW-2.0*DELTAX)/(maxX-xMin);
    /*return  DELTAX + (x-minX)*(scrW-2.0*DELTAX)/(maxX-minX);*/
}
/* Funzione per la conversione di una coordinata y nel piano cartesiano in
   una coordinata y di schermo. Usa le variabili globali
   - maxY: massima coordinata Y rilevata; si usa per riscalare la vista in modo da
           sfruttare al massimo lo schermo
   - minY: come sopra
   - scrH: altezza in pixel dello schermo
   y : => coordinata da trasformare
   return: coordinata a schermo
*/
int yToScreen (float y){
    float yMin = min (-1.0, minY);
    return  scrH - DELTAY - (y-yMin)*(scrH-2.0*DELTAY)/(maxY-yMin);
    /* le coordinate schermo sono ribaltate rispetto a quelle cartesiane:
       lo 0 è in alto e il massimo è in basso, quindi cambio di segno
    return  scrH - DELTAY - (y-minY)*(scrH-2.0*DELTAY)/(maxY-minY);*/
}
/* Procedura per il disegno di un punto in coordinate cartesiane
   Effettua la conversione in coordnate schermo e poi fa una "crocetta"
   per renderlo più visibile.
   a: => coordinata cartesiana x
   b: => coordinata cartesiana y
   color: => indice del colore di cui voglio il punto
*/
void disegnaPunto(float a, float b, int color){
    int x = xToScreen(a);
    int y = yToScreen(b);
    /* salva il colore in uso per ripristinarlo alla fine */
    int oldColor = getcolor();
    setcolor (color);
    putpixel (x-1, y-1, color);
    putpixel (x+1, y-1, color);
    putpixel (x  , y  , color);
    putpixel (x-1, y+1, color);
    putpixel (x+1, y+1, color);
    /* ripristina il colore precedente */
    setcolor (oldColor);
}
/* Procedura per il disegno di un punto in coordinate cartesiane
   e il disegno di un testo a lui vicino.
   a: => coordinata cartesiana x
   b: => coordinata cartesiana y
   lbl: => stringa da visualizzare
   color: => indice del colore di cui voglio il punto
*/
void disegnaPuntoELabel (float a, float b, char* lbl, int color){
    /* converto le coordinate cartesiane in punti schermo */
    int x = xToScreen(a);
    int y = yToScreen(b);
    /* salva il colore in uso per ripristinarlo alla fine */
    int oldColor = getcolor();
    /* disegna il punto */
    disegnaPunto (a, b, color);
    setcolor (color);
    /* disegna il testo */
    outtextxy (x+5, y+5, lbl);
    /* ripristina il colore precedente */
    setcolor (oldColor);
}
/* Procedura per il disegno di una retta nella forma y=ax+b
   Calcola il valore delle y nei punti di minima e massima x
   a: => coefficiente angolare della retta
   b: => termine noto
   color: => indice del colore di cui voglio la retta
*/
void disegnaRetta(float a, float b, int color){
    float y0 = a*minX+b;
    float y1 = a*maxX+b;

    float y0Scr = yToScreen(  y0);
    float y1Scr = yToScreen(  y1);
    float x0Scr = xToScreen(minX);
    float x1Scr = xToScreen(maxX);
    /* salva il colore in uso per ripristinarlo alla fine */
    int oldColor = getcolor();
    setcolor (color);

    line (x0Scr, y0Scr, x1Scr, y1Scr);
/*    line (xToScreen(minX), yToScreen(  y0),
          xToScreen(maxX), yToScreen(  y1));
                                         */
    setcolor (oldColor);
}
/* Procedura per il disegno degli assi cartesiani
   Disegna un asse solo se e per la parte compresa nella porzione di piano
   che viene usata per il disegno, ovvero quella compresa in (minX, minY)
   (maxX, maxY).
*/
void disegnaAssi (){
    /* salva il colore in uso per ripristinarlo alla fine */
    int oldColor = getcolor();
    setcolor(COLOREASSI);
    /* disegna l'asse x: da minX a maxX su y=0 */
    if (0>=minY&&0<=maxY){
        line (xToScreen(minX), yToScreen(   0),
              xToScreen(maxX), yToScreen(   0));
    }
    /* disegna l'asse y: da minY a maxY su x=0 */
    if (0>=minX&&0<=maxX){
        line (xToScreen(   0), yToScreen(minY),
              xToScreen(   0), yToScreen(maxY));
    }
    disegnaPuntoELabel(0, 0, "O", COLOREASSI);
    setcolor(oldColor);
}
/* Procedura per il disegno di una voce della legenda
   Effettua il disegno di unal inea del colore fissato
   e poi vi disegna accanto l'etichetta.
   idx: => posizione dall'alto in legenda
   color: => colore desiderato
   label: => etichetta da scrivere
*/
void disegnaVoceLegenda (int idx, int color, char* label){
    int vSpace = 18;
    int yText= 2+ (1+idx)*vSpace;
    int yLine= 4+ yText;
    /* salva il colore in uso per ripristinarlo alla fine */
    int oldColor = getcolor();
    setcolor(color);
    /* disegna la linea del colore prefissato */
    line (2, yLine, 19, yLine);
    /* scrive accanto l'etichetta */
    outtextxy( 24, yText, label);
    /* ripristina il colore precedente */
    setcolor(oldColor);
}
/* Procedura per il disegno della legenda
   Effettua il disegno di un rettangolo contenitore e poi
   disegna le varie voci.
*/
void disegnaLegenda (){
    setcolor(RED);
    rectangle (0,0, 200, 110);
    outtextxy( 2, 2, "Legenda:");
    disegnaVoceLegenda (0, COLORERETTARICAVI, "Retta_dei_ricavi");
    disegnaVoceLegenda (1, COLORERETTACOSTI, "Retta_dei_costi");
    disegnaVoceLegenda (2, COLORERETTAREGRESSIONE, "Retta_di_regressione");
    disegnaVoceLegenda (3, COLOREDATI, "Dati_rilevati");
    disegnaVoceLegenda (4, COLOREPUNTOMEDIO, "Punto_medio");
}
/* Procedura per il disegno di tutti punti rilevati
   p: => vettore dei punti da disegnare
   n: => numero di punti da disegnare
*/
void disegnaDatiRilevati (punto* p, int n){
    int i;
    for (i=0; i<n; i++){
        disegnaPunto(p[i].x, p[i].y, COLOREDATI);
    }
}

/* Procedura per il disegno del punto medio.
   Disegna il punto e scrive accanto Pm
   mediaX: => sommatoria dei valori delle x dei dati rilevati
   mediaY: => sommatoria dei valori delle y dei dati rilevati
   n: => numero dei dati rilevati
*/
void disegnaPuntoMedio (float mediaX, float mediaY, int n){
    float x = mediaX/(n);
    float y = mediaY/(n);
    float xScr = xToScreen(x);
    float yScr = yToScreen(y);
    disegnaPuntoELabel (xScr, yScr, "Pm", COLOREPUNTOMEDIO);
}

/* Procedura per il disegno della retta di regressione
   mediaX: => sommatoria dei valori delle x dei dati rilevati
   mediaY: => sommatoria dei valori delle y dei dati rilevati
   varX:   => varianza
   covarXY: => covarianza
   n: => numero dei dati rilevati
*/
void disegnaRettaDiRegressione (float mediaX, float mediaY, float varX, float covarXY, int n){
    float a = covarXY/varX;
    float b = mediaY/(n) -a*mediaX/(n);
    disegnaRetta (a, b, COLORERETTAREGRESSIONE);
}
/* Procedura per il disegno della retta dei costi fissi
   varX:   => varianza
   covarXY: => covarianza
   n: => numero dei dati rilevati
*/
void disegnaRettaDeiCostiFissi (float mediaX, float mediaY, float varX, float covarXY, int n){
    float a = 0.0;
    float b = mediaY/(n) - covarXY/varX*mediaX/(n);
    disegnaRetta (a, b, COLORERETTACOSTI);
}
/* Procedura per il disegno della retta dei ricavi totali
   prezzo: => prezzo di vendita
*/
void disegnaRettaDeiRicaviTotali (float prezzo){
    float a = prezzo;
    float b = 0.0;
    disegnaRetta (a, b, COLORERETTARICAVI);
}
/* Procedura per il calcolo del punto di equilibrio.
  Calcolato come punto di intersezione tra la retta di regressione
  e la retta dei costi totali.
   x: <= coordinata cartesiana del punto di equilibrio
   y: <= coordinata cartesiana del punto di equilibrio
   mediaX: => sommatoria dei valori delle x dei dati rilevati
   mediaY: => sommatoria dei valori delle y dei dati rilevati
   varX:   => varianza
   covarXY: => covarianza
*/
void CalcolaPuntoDiEquilibrio (float* x, float* y, float mediaX, float mediaY, float varX, float covarXY, int n, float prezzo){
    *x = (mediaX*covarXY-mediaY*varX )/((n)*(covarXY - prezzo*varX));
    *y = prezzo* *x;
}
/* Procedura per il disegno del punto di equilibrio.
   Disegna il punto e scrive accanto PEq
   x: => coordinata cartesiana del punto di equilibrio
   y: => coordinata cartesiana del punto di equilibrio
*/
void disegnaPuntoDiEquilibrio (float x, float y){
    disegnaPuntoELabel (x, y, "PEq", COLOREPUNTOMEDIO);
}

int main () {
    /* Variabili per la gestione della grafica                   */
    int gdriver = DETECT; /* autodetect del driver grafico       */
    int gmode, errorcode; /* modalità grafica e codice di errore */
    int prezzo;
    punto dati [NUM_MAX_PUNTI];
    int numDati=0;
    int ancora=1;
    float x,y;
    float xEq,yEq;
    float varx, covarxy, mediax=0.0, mediay=0.0, mediaqx=0.0, prodxy=0.0;
    /* Acquisisci i dati dall'utente */
    do {
	ancora  = acquisisciDato (dati, numDati);
	x = dati[numDati].x;
	y = dati[numDati].y;
	numDati = numDati+1; /* Aggiorna il numero dei dati rilevati*/
	mediax=mediax+x;     /* Aggiorna la somma che concorrerà alla media dei valori x */
	mediay=mediay+y;     /* Aggiorna la somma che concorrerà alla media dei valori y */
	mediaqx=mediaqx+x*x; /* */
	prodxy=prodxy+x*y;   /* */
        /* Aggiorna la "finestra" massima dei valori rilevati */
	if (x<minX){minX=x;}
	else if (x>maxX){maxX=x;}
	if (y<minY){minY=y;}
	else if (y>maxY){maxY=y;}
    }while(ancora);
    printf("Inserire il prezzo di vendita\n");
    scanf("%d",&prezzo);
    /* Effettua i calcoli delle varianze e delle covarianze */
    varx    = (mediaqx/(numDati))-(mediax*mediax/((numDati)*(numDati)));
    covarxy = (prodxy /(numDati))-(mediax*mediay/((numDati)*(numDati)));
    CalcolaPuntoDiEquilibrio (&xEq, &yEq, mediax, mediay, varx, covarxy, numDati, prezzo);
    /* Aggiorna la "finestra" massima dei valori rilevati */
    if (xEq<minX){minX=xEq;}
    else if (xEq>maxX){maxX=xEq;}
    if (yEq<minY){minY=yEq;}
    else if (yEq>maxY){maxY=yEq;}
    /*Stampa le equazioni*/
    printf("Punto medio: (%.2f,%.2f)\n",mediax/(numDati),mediay/(numDati));
    printf("Retta di regressione (costi totali): y=%.2f + %.2f (x- %.2f)\n",mediay/(numDati),covarxy/varx,mediax/(numDati));
    printf("Retta dei costi fissi: y= %.2f\n",covarxy/varx);
    printf("Retta dei ricavi totali: y=%d x\n",prezzo);
    printf("Punto di equilibrio: (%.2f,%.2f)\n",xEq, yEq);
    printf("Premi un tasto per terminare\n");
    getch();

    /* Inizializzazione dei driver grafici                       */
    initgraph(&gdriver, &gmode, "");
    /* Leggi i risultati dell'inizializzazione                   */
    errorcode = graphresult();
    if (errorcode != grOk) { /* C'è stato un errore              */
	printf("Errore grafico: %s\n", grapherrormsg(errorcode));
	printf("Premi un qualsiasi tasto per continuare:");
	getch();
	exit(1);
    }
    /* Acquisisci i dati della risoluzione dello schermo         */
    scrW = getmaxx();
    scrH = getmaxy();

    /* Disegna i risultati */
    cleardevice();
    setbkcolor (COLORESFONDO);
    disegnaAssi ();
    disegnaDatiRilevati (dati, numDati);
    disegnaRettaDiRegressione (mediax, mediay, varx, covarxy, numDati);
    disegnaRettaDeiCostiFissi (mediax, mediay, varx, covarxy, numDati);
    disegnaPuntoMedio (mediax, mediay, numDati);
    disegnaLegenda ();
    getch();

    cleardevice();
    setbkcolor (COLORESFONDO);
    disegnaAssi ();
    disegnaRettaDiRegressione (mediax, mediay, varx, covarxy, numDati);
    disegnaRettaDeiRicaviTotali (prezzo);
    disegnaPuntoDiEquilibrio (xEq, yEq);
    disegnaLegenda ();
    getch();

    cleardevice();
    /*Rilascia il driver grafico                                 */
    closegraph();
    return 0;
}
